# UniMCI (Universal MPI Correctness Interface)
# Copyright (c) 2008-2009, ZIH, TU Dresden, Federal Republic of Germany
#
# See the file LICENSE.txt in the package base directory for details
#
# @file CreatePnMPIModule.cmake
#       Creates a P^nMPI module for UniMCI, it provides an extremly easy 
#       access to checking functionality.
#
# TODO: currently not activated per default, build configuration not 
#       completly automated.
#
# @author Tobias Hilbrich
# @date 27.05.2009

#===================================
# Generator Macro 
#===================================
#Macro to MPI wrappers for the P^nMPI module
MACRO (UNIMCI_GENERATE_PnMPI_WRAPPERS infile outfile)
  FILE (READ ${infile} CONTENT)
  STRING (REPLACE "\n" "<BREAK><BREAK>" CONTENT ${CONTENT})

  #Loop over all the MPI calls
  STRING (REGEX MATCH "(^|<BREAK>)[a-zA-Z0-9]*[ ]MPI[^<>]*($|<BREAK>)" mpicall "${CONTENT}")

  WHILE (mpicall MATCHES "MPI" )

    #remove the found string from content
    STRING (REPLACE "${mpicall}" "" CONTENT ${CONTENT})
  
    #Read name of function
    STRING (REGEX MATCH "(^|<BREAK>)[a-zA-Z0-9]*[ ]MPI_[^(]*" mpifunction ${mpicall})
    STRING (REGEX REPLACE "(^|<BREAK>)[a-zA-Z0-9]*[ ]" "" mpifunction ${mpifunction})
    STRING (REPLACE "<BREAK>" "" header ${mpicall})
    
    #Remove the types from the function
    SET(REGEXES "([(])[^,]*[ ][^,]*(,)" "(,)[^,]*[ ][^,]*([)])" "([,])[^,]*[ ][^,]*(,)" "([(])[^,]*[ ][^,]*([)])")
    
    FOREACH(TESTEX ${REGEXES})
        WHILE (mpicall MATCHES ${TESTEX})
            #Determine the part that matches and the begin end charcter ",|(|)" 
            STRING (REGEX MATCH ${TESTEX} myMatch ${mpicall})
            STRING (REGEX REPLACE ${TESTEX} "\\1" myBegin ${myMatch})
            STRING (REGEX REPLACE ${TESTEX} "\\2" myEnd ${myMatch}) 
            
            ## myMatch = "[,|(]TEXT[,|)]" TEXT=type1 type2 ... typeN argument
            ## get last word of TEXT, remove any "*" characters !
            STRING (REGEX MATCH "[^ ]*[ ]*[${myEnd}]$" myArgument ${myMatch})        
            STRING (REGEX REPLACE "[ ]*[${myEnd}]$" "" myArgument ${myArgument})
            STRING (REPLACE "*" "" myArgument ${myArgument})
            STRING (REGEX REPLACE "[^a-zA-Z0-9_-].*[^a-zA-Z0-9_-]" "" myArgument ${myArgument})
            
            #Replace types and argument with argument
            STRING (REPLACE ${myMatch} "${myBegin}${myArgument}${myEnd}" mpicall ${mpicall})
            
            #MESSAGE ("${myMatch}: ${myBegin}${myArgument}${myEnd}")
        ENDWHILE (mpicall MATCHES ${TESTEX})
    ENDFOREACH(TESTEX ${REGEXES})
    
    ##Generate the two unimci check calls
    #kill return value
    STRING (REGEX REPLACE "((^|<BREAK>))[a-zA-Z0-9]*[ ]" "\\1 " mpicall ${mpicall})
    SET (unimcicall ${mpicall})
    
    #Add sourcefile name, linenumber, and eventDescriptor to the end of the file
    IF (unimcicall MATCHES "MPI_.*[(]void[)]")
      #special rule for calls with no arguments
      STRING (REGEX REPLACE 
             "void[)][ ]*($|<BREAK>)" 
             "\"\", 0, 0)\\1" 
             unimcicall ${unimcicall})
      STRING (REGEX REPLACE 
             "void[)][ ]*($|<BREAK>)" 
             ")\\1" 
             mpicall ${mpicall})
    ELSE (unimcicall MATCHES "MPI_.*[(]void[)]")
      #normal rule
      STRING (REGEX REPLACE 
             "[)][ ]*($|<BREAK>)" 
             " , \"\", 0, 0)\\1" 
             unimcicall ${unimcicall})
    ENDIF (unimcicall MATCHES "MPI_.*[(]void[)]")
    
    #For pre/post: MPI_ -> UNIMCI_check___MPI_
    #For mpicall: MPI_ -> PMPI_
    STRING (REGEX REPLACE "((^|<BREAK>)[a-zA-Z0-9]*[ ])MPI_" "\\1UNIMCI_check___MPI_" unimcicall ${unimcicall})
    STRING (REGEX REPLACE "((^|<BREAK>)[a-zA-Z0-9]*[ ])MPI_" "\\1PMPI_" mpicall ${mpicall})
                
    #create two versions, one for pre call and one for post call
    #pre: mpicall, post: postcall
    SET (postcall ${unimcicall})
    STRING (REGEX REPLACE "___" "_pre__" unimcicall ${unimcicall})
    STRING (REGEX REPLACE "___" "_post__" postcall ${postcall})   
  
    #clean the mpicall
    STRING (REPLACE "<BREAK>" "" unimcicall ${unimcicall})
    STRING (REPLACE "<BREAK>" "" postcall ${postcall})
    STRING (REPLACE "<BREAK>" "" mpicall ${mpicall})
    
    FILE (APPEND ${outfile}
          "/* ${mpifunction} */\n"
          "${header}\n"
          "{\n" 
          "  int ret;\n"
          " ${unimcicall};\n" 
          "  UniMCI_PnMPI_handle_errors();\n"
          "  ret =${mpicall};\n"
          " ${postcall};\n"
          "  UniMCI_PnMPI_handle_errors();\n"
          "  return ret;\n"
          "}\n\n")
     
    #next iteration 
    STRING (REGEX MATCH "(^|<BREAK>)[a-zA-Z0-9]*[ ]MPI[^<>]*($|<BREAK>)" mpicall "${CONTENT}")
  ENDWHILE (mpicall MATCHES "MPI" )
ENDMACRO (UNIMCI_GENERATE_PnMPI_WRAPPERS infile outfile)

#===================================
# Actual CMake stuff 
#===================================
##TODO: Once a P^nMPI package is available we may enable this
#OPTION(UNIMCI_BUILD_PNMPI_MODULE "If active a P^nMPI module for UniMCI is built." OFF)
#MARK_AS_ADVANCED(UNIMCI_BUILD_PNMPI_MODULE)

IF (UNIMCI_BUILD_PNMPI_MODULE)
    

    #Generate the source file
    SET (INFILE "${PROJECT_SOURCE_DIR}/src/include/mpi-1-functions.h")
    SET (OUTFILE "${PROJECT_BINARY_DIR}/generated/c/pnmpi_module.c")
    CONFIGURE_FILE("${PROJECT_SOURCE_DIR}/src/c/pnmpi_module_core.c.in" "${OUTFILE}" COPYONLY)
    UNIMCI_GENERATE_PnMPI_WRAPPERS (${INFILE} ${OUTFILE})
    FILE (APPEND ${OUTFILE} "")
    
    #Add librarie/include dirs (P^nMPI)
    SET (PNMPI_LIB_DIR  "/home/hilbrich/local/lib" 
         CACHE PATH     "The path for the P^nMPI main library.")
    SET (PNMPI_LIBS     "-lpnmpi"                  
         CACHE STRING   "The P^nMPI Library.")
    SET (PNMPI_PATCHER  "/home/hilbrich/pnmpi-1.3/patch/patch"
         CACHE FILEPATH "The P^nMPI library patcher executable.")
    SET (PNMPI_INC_DIR  "/home/hilbrich/pnmpi-1.3/src"
         CACHE PATH     "The directory that contains the P^nMPI headers.")
    INCLUDE_DIRECTORIES("${PNMPI_INC_DIR}")
    
    #Create the target
    #SET(CMAKE_VERBOSE_MAKEFILE ON) ##TODO REMOVE
    SET(CMAKE_C_FLAGS ${UNIMCI_MPI_C_FLAGS}) # "-g")
    SET(CMAKE_C_COMPILER ${UNIMCI_MPI_C_COMPILER})
    LINK_DIRECTORIES("${PROJECT_BINARY_DIR}/lib" ${UNIMCI_MPI_C_LIB_DIR_LIST})
        
    ADD_LIBRARY (unimci_module MODULE "${PROJECT_BINARY_DIR}/generated/c/pnmpi_module.c")
    
    #Set the target library dependencies
    STRING(REPLACE "unimci-util" "unimci-util-shared" lib_list "${UNIMCI_MPI_C_LIB_LIST}")
    TARGET_LINK_LIBRARIES (unimci_module ${lib_list})
    
    #Patch and install
    SET_TARGET_PROPERTIES (unimci_module
    PROPERTIES 
        VERSION 1.0.0 #TODO
        SOVERSION 1 #TODO
    )    
    
    INSTALL(TARGETS unimci_module
            PERMISSIONS 
                OWNER_READ OWNER_WRITE OWNER_EXECUTE 
                GROUP_EXECUTE GROUP_READ 
                WORLD_EXECUTE
            RUNTIME DESTINATION bin
            LIBRARY DESTINATION lib
            ARCHIVE DESTINATION lib
            )    
    
    MUST_MAC_GET_TARGET_MODULE_NAME (name unimci_module 1.0.0)
    MUST_MAC_ADD_PNMPI_PATCH (${name} ${PNMPI_PATCHER})   
ENDIF (UNIMCI_BUILD_PNMPI_MODULE)
 