# UniMCI (Universal MPI Correctness Interface)
# Copyright (c) 2008-2009, ZIH, TU Dresden, Federal Republic of Germany
#
# See the file LICENSE.txt in the package base directory for details

# @file HelperMacros.cmake
#       UniMCI helper macros, partially based on Marmot macros from Blasius Czink. 
#
# @author Tobias Hilbrich, Blasius Czink
# @date 26.11.2008

MACRO (BC_LIST_CONTAINS var value)
  SET (${var})
  FOREACH (value2 ${ARGN})
    STRING (TOUPPER ${value}  str1)
    STRING (TOUPPER ${value2} str2)
    IF (${str1} STREQUAL ${str2})
      SET (${var} TRUE)
    ENDIF (${str1} STREQUAL ${str2})
  ENDFOREACH (value2)
ENDMACRO (BC_LIST_CONTAINS)

#====================================
# Macro from UniMCI
# Does:
#   if <elem> is the ith element in <listA>
#   it returns the ith element in <listB>
#   ! Limited to strings 
#   the return is stored in <return>
#====================================
MACRO (UNIMCI_MAC_STRING_CORRESPOND elem listA listB return)
  SET (I a) #No number operations in CMake :( (So we count with a's)
  SET (FOUND_IT FALSE)
  FOREACH (elemA ${listA})
    IF (${elem} STREQUAL ${elemA})
      SET (FOUND_IT TRUE)    
    ENDIF (${elem} STREQUAL ${elemA})
    
    IF (NOT FOUND_IT)
        STRING (REGEX REPLACE "a$" "aa" I ${I})
    ENDIF (NOT FOUND_IT)
  ENDFOREACH (elemA)
  
  IF (FOUND_IT)
    FOREACH (elemB ${listB})
      IF (FOUND_IT)
        IF (${I} STREQUAL "a")
          SET (${return} ${elemB})
          SET (FOUND_IT FALSE)
        ENDIF (${I} STREQUAL "a")
      ENDIF (FOUND_IT)
      
      IF (FOUND_IT)
        STRING (REGEX REPLACE "aa$" "a" I ${I})
      ENDIF (FOUND_IT)
    ENDFOREACH (elemB)
  ENDIF (FOUND_IT)
ENDMACRO (UNIMCI_MAC_STRING_CORRESPOND) 

#====================================
# Macro from UniMCI
# Does:
#   Gets a lib and lib-path (<libs>) and 
#   string and parses it for its libary 
#   pathes. Each path is appended to
#   <lib_dir_list>, which gets initialized
#   by an empty list.
#====================================
MACRO (UNIMCI_MAC_GET_LIB_DIR_LIST libs lib_dir_list)
  SET (${lib_dir_list})
  SET (LIBS_COPY ${libs})
  
  #TODO Likely specific to Linux, enhance for other OS's if necessary
  STRING (REGEX MATCH "[ ]-L[^ ]*" A_LIB_DIR " ${LIBS_COPY} ")
  
  WHILE (A_LIB_DIR)
    #extract lib dir
    STRING (REPLACE ${A_LIB_DIR} "" LIBS_COPY " ${LIBS_COPY} ")
    STRING (REGEX REPLACE "^.* -L" "" A_LIB_DIR "${A_LIB_DIR}")
  
    IF (${lib_dir_list})
      SET (${lib_dir_list} "${${lib_dir_list}};${A_LIB_DIR}")
    ELSE (${lib_dir_list})
      SET (${lib_dir_list} "${A_LIB_DIR}")
    ENDIF (${lib_dir_list})
    
    #next iteration
    STRING (REGEX MATCH "[ ]-L[^ ]*" A_LIB_DIR " ${LIBS_COPY} ")
  ENDWHILE (A_LIB_DIR)  
  
ENDMACRO (UNIMCI_MAC_GET_LIB_DIR_LIST) 

#====================================
# Macro from UniMCI
# Does:
#   Gets a lib and lib-path (<libs>) and 
#   string and parses it for its libaries 
#   Each library is appended to
#   <lib_dir_list>, which gets initialized
#   by an empty list.
#====================================
MACRO (UNIMCI_MAC_GET_LIB_LIST libs lib_list)
  #Init
  SET (${lib_list})
  SET (LIBS_COPY ${libs})
  
  #TODO Likely specific to Linux, enhance for other OS's if necessary
  #Ask for the lib dirs so we can throw them away
  UNIMCI_MAC_GET_LIB_DIR_LIST (${LIBS_COPY} LIB_DIR_LIST)
  
  FOREACH (LIB_DIR ${LIB_DIR_LIST})
    STRING (REGEX REPLACE "-L${LIB_DIR}" "" LIBS_COPY ${LIBS_COPY})
  ENDFOREACH (LIB_DIR)
  
  STRING (REGEX REPLACE " [ ]+" " " LIBS_COPY ${LIBS_COPY})
  STRING (REGEX REPLACE "^[ ]+" "" LIBS_COPY " ${LIBS_COPY}")
  STRING (REGEX REPLACE "[ ]+$" "" LIBS_COPY "${LIBS_COPY} ")
  
  #Get the libs and put them into a list
  STRING (REGEX MATCH "[ ]*[^ ]*( |$)" A_LIB "${LIBS_COPY}")
  
  WHILE (A_LIB AND LIBS_COPY)
    #remove from the libs copy
    STRING (REPLACE "${A_LIB}" "" LIBS_COPY "${LIBS_COPY}")
    STRING (REPLACE " " "" A_LIB "${A_LIB}")
  
    IF (${lib_list})
      SET (${lib_list} "${${lib_list}};${A_LIB}")
    ELSE (${lib_list})
      SET (${lib_list} "${A_LIB}")
    ENDIF (${lib_list})
    
    #next iteration
    IF (LIBS_COPY)
        STRING (REGEX MATCH "[ ]*[^ ]*( |$)" A_LIB "${LIBS_COPY}")
    ENDIF (LIBS_COPY)
  ENDWHILE (A_LIB AND LIBS_COPY)  
ENDMACRO (UNIMCI_MAC_GET_LIB_LIST)

#====================================
# Macro from UniMCI
# Does:
#   Compares a Marmot version to a given version
#====================================
MACRO (UNIMCI_MAC_MARMOT_VERSION_GE version_string test_version result)
    
    STRING (REGEX MATCH "^[^.]*" mayor "${version_string}")
    STRING (REGEX MATCH "[^.]*$" patch "${version_string}")
    STRING (REGEX MATCH "[.][^.]*[.]" minor "${version_string}")
    STRING (REPLACE "." "" minor "${minor}")
    
    STRING (REGEX MATCH "^[^.]*" test_mayor "${test_version}")
    STRING (REGEX MATCH "[^.]*$" test_patch "${test_version}")
    STRING (REGEX MATCH "[.][^.]*[.]" test_minor "${test_version}")
    STRING (REPLACE "." "" test_minor "${test_minor}")
    
    MATH(EXPR compare "${mayor}*1000000 + ${minor}*100 + ${patch} - ${test_mayor}*1000000 - ${test_minor}*100 - ${test_patch}")
    IF (compare MATCHES "[-]")
        SET (${result} FALSE)
    ELSE (compare MATCHES "[-]")
        SET (${result} TRUE)
    ENDIF (compare MATCHES "[-]")
             
ENDMACRO (UNIMCI_MAC_MARMOT_VERSION_GE)

#=========================================================
# Macro from MUST
# Does:
#   Gets the name of a library, Gets the path to the 
#   P^nMPI patcher, Creates a CMake script in the binary
#   directory which executes the patcher for the given 
#   library. The result is placed in the P^nMPI library 
#   directory. The script is added as a install script.
#   (executed at install time)
#=========================================================
MACRO (MUST_MAC_ADD_PNMPI_PATCH lib patcher)
    #TODO: test whether this works with Windows pathes (spaces and such)
    FILE (WRITE ${PROJECT_BINARY_DIR}/install-scripts/patch-${lib}.cmake 
        "MESSAGE (\"Patching ${lib}\")\n"
        "EXECUTE_PROCESS (COMMAND ${patcher}"
        "   ${PROJECT_BINARY_DIR}/lib/${lib} ${CMAKE_INSTALL_PREFIX}/lib/${lib}"
        "   WORKING_DIRECTORY ${PROJECT_BINARY_DIR}/install-scripts"
        "   RESULT_VARIABLE result"
        "   OUTPUT_VARIABLE output"
        "   ERROR_VARIABLE output)\n"
        "IF (result)\n"
        "   MESSAGE (FATAL_ERROR \${output})\n"
        "ENDIF (result)\n"
        )
    INSTALL (SCRIPT ${PROJECT_BINARY_DIR}/install-scripts/patch-${lib}.cmake)
ENDMACRO (MUST_MAC_ADD_PNMPI_PATCH)

#========================================================
# Macro from MUST
# Does:
#   Saves the file name for a target module in <name>.
#   This is calculated from <targetname>, name of the 
#   module target and <version> the full version of the
#   module.
# TODO:
#   Likely needs extensions for other platforms like Win.
#=========================================================   
MACRO (MUST_MAC_GET_TARGET_MODULE_NAME name targetname version)
    SET (name ${CMAKE_SHARED_MODULE_PREFIX}${targetname}${CMAKE_SHARED_MODULE_SUFFIX}.${version})
ENDMACRO (MUST_MAC_GET_TARGET_MODULE_NAME)
