# UniMCI (Universal MPI Correctness Interface)
# Copyright (c) 2008-2009, ZIH, TU Dresden, Federal Republic of Germany
#
# See the file LICENSE.txt in the package base directory for details

# @file MARMOTGenerate.cmake
#       Generator (if needed) for Marmot specific UniMCI implementations. 
#
# @author Tobias Hilbrich
# @date 03.12.2008

#===================================
# Processing Macro 
#===================================
#Macro to process a file with mpi function prototyes
#It creates Marmot specific implementations/redefinitions
#for each MPI function
MACRO (UNIMCI_GENERATE_MARMOT_HEADER infile outfile implemented)
  FILE (READ ${infile} CONTENT)
  STRING (REPLACE "\n" "<BREAK><BREAK>" CONTENT ${CONTENT})
  STRING (REPLACE "\n" "<BREAK>" my_implemented "\n${implemented}\n")

  #Loop over all the MPI calls
  STRING (REGEX MATCH "(^|<BREAK>)[a-zA-Z0-9]*[ ]MPI[^<>]*($|<BREAK>)" mpicall "${CONTENT}")

  WHILE (mpicall MATCHES "MPI" )

    #remove the found string from content
    STRING (REPLACE "${mpicall}" "" CONTENT ${CONTENT})
  
    #See whether the function is implemented
    STRING (REGEX MATCH "(^|<BREAK>)[a-zA-Z0-9]*[ ]MPI_[^(]*" mpifunction ${mpicall})
    STRING (REGEX REPLACE "(^|<BREAK>)[a-zA-Z0-9]*[ ]" "" mpifunction ${mpifunction})
    STRING (REGEX MATCH "<BREAK>${mpifunction}<BREAK>" match ${my_implemented})
      
    IF (match MATCHES "${mpifunction}")
      #replace 'MPI_' by 'MARMOT_' (for function name)
      STRING (REGEX REPLACE "((^|<BREAK>)[a-zA-Z0-9]*[ ])MPI_" "\\1UNIMCI_check___MPI_" mpicall ${mpicall})
    
      #remove arguments and return value
      STRING (REGEX REPLACE "(^|<BREAK>)[a-zA-Z0-9]*[ ]" "" mpicall ${mpicall})
      STRING (REGEX REPLACE "(UNIMCI_check___MPI_[^(]*).*" "\\1" mpicall ${mpicall})
          
      #create two versions, one for pre call and one for post call
      #pre: mpicall, post: postcall
      SET (postcall ${mpicall})
      STRING (REGEX REPLACE "___" "_pre__" mpicall ${mpicall})
      STRING (REGEX REPLACE "___" "_post__" postcall ${postcall})   
  
      #clean the mpicall
      STRING (REPLACE "<BREAK>" "" mpicall ${mpicall})
      STRING (REPLACE "<BREAK>" "" postcall ${postcall})
    
      #create the second Marmot version
      STRING (REGEX REPLACE "UNIMCI_" "MARMOT_" marmotcall ${mpicall})
      STRING (REGEX REPLACE "UNIMCI_" "MARMOT_" marmotpostcall ${postcall})
    
      FILE (APPEND ${outfile} 
            "#define ${mpicall} ${marmotcall}\n"
            "#define ${postcall} ${marmotpostcall}\n\n")
    ENDIF (match MATCHES "${mpifunction}")
 
    #next iteration 
    STRING (REGEX MATCH "(^|<BREAK>)[a-zA-Z0-9]*[ ]MPI[^<>]*($|<BREAK>)" mpicall "${CONTENT}")
  ENDWHILE (mpicall MATCHES "MPI" )
ENDMACRO (UNIMCI_GENERATE_MARMOT_HEADER infile outfile implemented)

#===================================
# Processing Macro 
#===================================
#Macro to create dummy implementations for MPI 
#calls not handled by Marmot
MACRO (UNIMCI_GENERATE_MARMOT_DUMMY infile outfile implemented)
  FILE (READ ${infile} CONTENT)
  STRING (REPLACE "\n" "<BREAK><BREAK>" CONTENT ${CONTENT})
  STRING (REPLACE "\n" "<BREAK>" my_implemented "\n${implemented}\n")

  #Loop over all the MPI calls
  STRING (REGEX MATCH "(^|<BREAK>)[a-zA-Z0-9]*[ ]MPI[^<>]*($|<BREAK>)" mpicall "${CONTENT}")

  WHILE (mpicall MATCHES "MPI" )

    #remove the found string from content
    STRING (REPLACE "${mpicall}" "" CONTENT ${CONTENT})
  
    #See whether the function is implemented
    STRING (REGEX MATCH "(^|<BREAK>)[a-zA-Z0-9]*[ ]MPI_[^(]*" mpifunction ${mpicall})
    STRING (REGEX REPLACE "(^|<BREAK>)[a-zA-Z0-9]*[ ]" "" mpifunction ${mpifunction})
    STRING (REGEX MATCH "<BREAK>[ ]*${mpifunction}[ ]*<BREAK>" match ${my_implemented})
    
    IF (NOT match MATCHES "${mpifunction}")
      #replace 'MPI_' by 'MARMOT_' (for function name)
      STRING (REGEX REPLACE "((^|<BREAK>)[a-zA-Z0-9]*[ ])MPI_" "\\1UNIMCI_check___MPI_" mpicall ${mpicall})
    
      #kill return value
      STRING (REGEX REPLACE "((^|<BREAK>))[a-zA-Z0-9]*[ ]" "\\1void " mpicall ${mpicall})
      
      #Add sourcefile name, linenumber, and eventDescriptor to the end of the file
      IF (mpicall MATCHES "MPI_.*[(]void[)]")
        #special rule for calls with no arguments
        STRING (REGEX REPLACE 
               "void[)][ ]*($|<BREAK>)" 
               "const char* unimci_file, unsigned int unimci_line, unsigned long long unimci_eventDescriptor)\\1" 
               mpicall ${mpicall})
      ELSE (mpicall MATCHES "MPI_.*[(]void[)]")
        #normal rule
        STRING (REGEX REPLACE 
               "[)][ ]*($|<BREAK>)" 
               " ,const char* unimci_file, unsigned int unimci_line, unsigned long long unimci_eventDescriptor)\\1" 
               mpicall ${mpicall})
      ENDIF (mpicall MATCHES "MPI_.*[(]void[)]")
                
      #create two versions, one for pre call and one for post call
      #pre: mpicall, post: postcall
      SET (postcall ${mpicall})
      STRING (REGEX REPLACE "___" "_pre__" mpicall ${mpicall})
      STRING (REGEX REPLACE "___" "_post__" postcall ${postcall})   
  
      #clean the mpicall
      STRING (REPLACE "<BREAK>" "" mpicall ${mpicall})
      STRING (REPLACE "<BREAK>" "" postcall ${postcall})
    
      FILE (APPEND ${outfile} 
            "${mpicall}\n" 
            "{}\n"
            "${postcall}\n"
            "{}\n\n")
    ENDIF (NOT match MATCHES "${mpifunction}")
 
    #next iteration 
    STRING (REGEX MATCH "(^|<BREAK>)[a-zA-Z0-9]*[ ]MPI[^<>]*($|<BREAK>)" mpicall "${CONTENT}")
  ENDWHILE (mpicall MATCHES "MPI" )
ENDMACRO (UNIMCI_GENERATE_MARMOT_DUMMY infile outfile implemented)

#==============================================
# Set the influencial build variables, we have:
#==============================================
# ${USED_MPI_CHECKER}_EXTRA_LIB_SRCS
# ${USED_MPI_CHECKER}_UNIMCI_HEADER_INCLUDE_PRE
# ${USED_MPI_CHECKER}_UNIMCI_HEADER_INCLUDE_POST
CONFIGURE_FILE ("${PROJECT_SOURCE_DIR}/src/c/marmot-core-impl.c.in" "${PROJECT_BINARY_DIR}/generated/c/marmot-core-impl.c")
SET (${USED_MPI_CHECKER}_EXTRA_LIB_SRCS 
     "${PROJECT_BINARY_DIR}/generated/c/marmot-core-impl.c"
     "${PROJECT_BINARY_DIR}/generated/c/marmot-mpi-1-dummies.c")
     #We will add mpi-2 dummies after a try compile ... (MPI might not support MPI-2)
SET (${USED_MPI_CHECKER}_UNIMCI_HEADER_INCLUDE_PRE "#include \"unimci-marmot-redef.h\"")
SET (${USED_MPI_CHECKER}_UNIMCI_HEADER_INCLUDE_POST "")
SET (${USED_MPI_CHECKER}_EXTRA_INSTALL_HEADERS "${PROJECT_BINARY_DIR}/generated/include/unimci-marmot-redef.h")

#==============================================
# Create a header that redefines the UniMCI MPI functions
#==============================================
#Open the output file
SET (OUTFILE "${PROJECT_BINARY_DIR}/generated/include/unimci-marmot-redef.h")
FILE (WRITE ${OUTFILE} 
      "/*\n"
      " * UniMCI (Universal MPI Correctness Interface)\n"
      " * Copyright (c) 2008-2009, ZIH, TU Dresden, Federal Republic of Germany\n"
      " *\n"
      " * See the file LICENSE.txt in the package base directory for details\n"
      " */\n\n"
      "/**\n"
      " * THIS FILE IS AUTOMATICALLY GENERATED,\n"
      " * ALL MODIFICATIONS WILL BE NON-PERMANENT!\n"
      " * \n"
      " * @file unimci-marmot-redef.h\n"
      " * Header file that:\n"
      " *  - defines UNIMCI MPI functions with a macro to\n"
      " *    redirect them to Marmot calls (if implemented)\n"
      " */\n\n"
      "\n"
      "#ifndef __UNIMCI_MARMOT_REDEF_H\n"
      "#define __UNIMCI_MARMOT_REDEF_H\n"
      "\n")
 
#Open the list of Marmot implemented functions
#TODO currently the implemented_functions.txt has to be in sync
#     with the Marmot installation. To overcome this Marmot should
#     install this file, so we can open it from Marmots installation.
#MARMOT_VERSION
IF (MARMOT_HAS_FULL_MPI2)
    #Marmot with full MPI2 support
    FILE (READ "${PROJECT_SOURCE_DIR}/src/include/marmot2.4_implemented_functions.txt" IMPLEMENTED)
ELSE (MARMOT_HAS_FULL_MPI2)
    #Marmot with no full (maybe partial, maybe none at all) MPI2 support
    FILE (READ "${PROJECT_SOURCE_DIR}/src/include/marmot_implemented_functions.txt" IMPLEMENTED)
ENDIF (MARMOT_HAS_FULL_MPI2)

#Issue the header generation for MPI-1 and MPI-2
SET (INFILE "${PROJECT_SOURCE_DIR}/src/include/mpi-1-functions.h")
UNIMCI_GENERATE_MARMOT_HEADER (${INFILE} ${OUTFILE} ${IMPLEMENTED})

FILE (APPEND ${OUTFILE} 
      "\n"
      "#ifdef UNIMCI_MPI2\n") 
SET (INFILE "${PROJECT_SOURCE_DIR}/src/include/mpi-2-functions.h")
UNIMCI_GENERATE_MARMOT_HEADER (${INFILE} ${OUTFILE} ${IMPLEMENTED})
FILE (APPEND ${OUTFILE} 
      "\n"
      "#endif /*UNIMCI_MPI2*/\n"
      "\n"
      "#endif /*__UNIMCI_MARMOT_REDEF_H*/\n")
      
#==============================================
# Create the dummy source files
#==============================================
#Open the output file for MPI-1
SET (OUTFILE "${PROJECT_BINARY_DIR}/generated/c/marmot-mpi-1-dummies.c")
FILE (WRITE ${OUTFILE} 
      "/*\n"
      " * UniMCI (Universal MPI Correctness Interface)\n"
      " * Copyright (c) 2008-2009, ZIH, TU Dresden, Federal Republic of Germany\n"
      " *\n"
      " * See the file LICENSE.txt in the package base directory for details\n"
      " */\n\n"
      "/**\n"
      " * THIS FILE IS AUTOMATICALLY GENERATED,\n"
      " * ALL MODIFICATIONS WILL BE NON-PERMANENT!\n"
      " * \n"
      " * @file marmot-mpi-1-dummies.c\n"
      " * Creates dummy implementations for all MPI calls\n"
      " * that are not implemented by Marmot.\n"
      " */\n\n"
      "\n"
      "#include <mpi.h>\n"
      "#include \"unimci-mpi-1.h\"\n"
      "\n")
 
#Issue the dummy generation for MPI-1
SET (INFILE "${PROJECT_SOURCE_DIR}/src/include/mpi-1-functions.h")
UNIMCI_GENERATE_MARMOT_DUMMY (${INFILE} ${OUTFILE} ${IMPLEMENTED})

#Open the output file for MPI-2
SET (OUTFILE "${PROJECT_BINARY_DIR}/generated/c/marmot-mpi-2-dummies.c")
FILE (WRITE ${OUTFILE} 
      "/*\n"
      " * UniMCI (Universal MPI Correctness Interface)\n"
      " * Copyright (c) 2008-2009, ZIH, TU Dresden, Federal Republic of Germany\n"
      " *\n"
      " * See the file LICENSE.txt in the package base directory for details\n"
      " */\n\n"
      "/**\n"
      " * THIS FILE IS AUTOMATICALLY GENERATED,\n"
      " * ALL MODIFICATIONS WILL BE NON-PERMANENT!\n"
      " * \n"
      " * @file marmot-mpi-2-dummies.c\n"
      " * Creates dummy implementations for all MPI calls\n"
      " * that are not implemented by Marmot.\n"
      " */\n\n"
      "\n"
      "#include <mpi.h>\n"
      "#include \"unimci-mpi-2.h\"\n"
      "\n")
 
#Issue the dummy generation for MPI-2
SET (INFILE "${PROJECT_SOURCE_DIR}/src/include/mpi-2-functions.h")
UNIMCI_GENERATE_MARMOT_DUMMY (${INFILE} ${OUTFILE} ${IMPLEMENTED})

#==============================================
# Test whether MPI-2 dummies work
#==============================================
#To test it we see whether we can at least compile and link a C application with Marmot
#Create a CMakeLists.txt file which will generate the test
IF (WIN32)
    STRING (REPLACE "\"" "\\\"" UNIMCI_MPI_C_FLAGS_TEMP ${UNIMCI_MPI_C_FLAGS})
ENDIF (WIN32)
FILE(WRITE ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp/CMakeLists.txt
  "CMAKE_MINIMUM_REQUIRED(VERSION 2.6)\n"
  "PROJECT(MPI2Test C)\n"
  "SET(CMAKE_C_FLAGS \"${UNIMCI_MPI_C_FLAGS_TEMP}\")\n"
  "SET(CMAKE_C_COMPILER \"${UNIMCI_MPI_C_COMPILER}\")\n"
  "SET(CMAKE_EXE_LINKER_FLAGS \"${UNIMCI_MPI_LD_FLAGS}\")\n"
  "LINK_DIRECTORIES("
  )

FOREACH (LIB ${UNIMCI_MPI_C_LIB_DIR_LIST})
   FILE (APPEND ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp/CMakeLists.txt
         "\"${LIB}\" ")
ENDFOREACH (LIB)

FILE (APPEND ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp/CMakeLists.txt
  ")\n"
  "LINK_LIBRARIES (${UNIMCI_MPI_C_LIB_LIST})\n"
  "INCLUDE_DIRECTORIES (\"${PROJECT_BINARY_DIR}/generated/include\")\n"
  "ADD_LIBRARY(MPI2Test \"${PROJECT_BINARY_DIR}/generated/c/marmot-mpi-2-dummies.c\")\n"
  )
                
# Use TRY_COMPILE to make the target "MPI2Test"
TRY_COMPILE(
  TEST_OK
  ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp
  ${PROJECT_BINARY_DIR}${CMAKE_FILES_DIRECTORY}/CMakeTmp
  MPI2Test
  OUTPUT_VARIABLE MY_OUTPUT)

IF (TEST_OK)
  #we can compile the MPI dummy source, so we cann add it to the targets
  SET (${USED_MPI_CHECKER}_EXTRA_LIB_SRCS 
     ${${USED_MPI_CHECKER}_EXTRA_LIB_SRCS}
     "${PROJECT_BINARY_DIR}/generated/c/marmot-mpi-2-dummies.c")
ENDIF (TEST_OK)
