/**
 *  @file
 *
 *  Very simple UniMCI test case.
 *
 *  It just calls
 *  - MPI_Init
 *  - MPI_Finalize
 *  + their UniMCI pre and post calls
 *
 *  @author Tobias Hilbrich
 *
 *  $ID$
 */


#include <mpi.h>
#include "unimci.h"
#include <stdio.h>

void check_unimci_return (void);
void print_unimci_message (UNIMCI_MSG *msg);

int main(int argc, char **argv)
{
    UNIMCI_check_pre__MPI_Init (&argc, &argv, __FILE__, __LINE__, 1);
    check_unimci_return();
    MPI_Init(&argc, &argv);
    UNIMCI_check_post__MPI_Init (&argc, &argv, __FILE__, __LINE__, 2);
    check_unimci_return();

    UNIMCI_check_pre__MPI_Finalize (__FILE__, __LINE__, 3);
    check_unimci_return();
    MPI_Finalize();
    UNIMCI_check_post__MPI_Finalize (__FILE__, __LINE__, 4);
    check_unimci_return();

    return 0;
}

void check_unimci_return (void)
{
    UNIMCI_MSG *msg;

    while (UNIMCI_has_msg ())
    {
        UNIMCI_pop_msg (&msg);

        print_unimci_message (msg);

        UNIMCI_msg_free (&msg);
    }
}

void print_unimci_message (UNIMCI_MSG *msg)
{
    printf ("%s\n",msg->strText);
}

/*EOF*/
