/**
 *  datatype.c
 *
 *  This is an MPI-Program with an error!
 *
 *  It just calls
 *          - MPI_Init
 *          - MPI_Comm_rank
 *          - MPI_Comm_size
 *          - MPI_Type_contiguous
 *          - MPI_Send
 *          - MPI_Recv
 *          - MPI_Type_free
 *
 * @author Tobias Hilbrich
 *
 */

#include <stdio.h>
#include <mpi.h>
#include "unimci.h"

void check_unimci_return (void);
void print_unimci_message (UNIMCI_MSG *msg);

int main(int argc, char **argv)
{
    int rank;
    int size;
    int sendBuf[4];
    int recvBuf[4];
    int MSG_TAG_1 = 1,
        MSG_TAG_2 = 2;

    MPI_Status status;

    MPI_Datatype cont2Int;
    MPI_Datatype cont4Int;

    UNIMCI_check_pre__MPI_Init (&argc, &argv, __FILE__, __LINE__, 1);
    check_unimci_return();
    MPI_Init(&argc, &argv);
    UNIMCI_check_post__MPI_Init (&argc, &argv, __FILE__, __LINE__, 2);
    check_unimci_return();

    UNIMCI_check_pre__MPI_Comm_rank(MPI_COMM_WORLD, &rank, __FILE__, __LINE__, 3);
    check_unimci_return();
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    UNIMCI_check_post__MPI_Comm_rank(MPI_COMM_WORLD, &rank, __FILE__, __LINE__, 4);
    check_unimci_return();

    UNIMCI_check_pre__MPI_Comm_size(MPI_COMM_WORLD, &size, __FILE__, __LINE__, 5);
    check_unimci_return();
    MPI_Comm_size(MPI_COMM_WORLD, &size);
    UNIMCI_check_post__MPI_Comm_size(MPI_COMM_WORLD, &size, __FILE__, __LINE__, 6);
    check_unimci_return();

    printf(" I am rank %d of %d PEs\n", rank, size);

    if (size < 2)
    {
        fprintf(stderr, " This program needs at least 2 PEs!\n");
    }
    else
    {
       sendBuf[0] = 1; sendBuf[1] = 2; sendBuf[2] = 3; sendBuf[3] = 4;

       //create two nice Datatype
       UNIMCI_check_pre__MPI_Type_contiguous(2, MPI_INT, &cont2Int, __FILE__, __LINE__, 7);
       check_unimci_return();
       MPI_Type_contiguous(2, MPI_INT, &cont2Int);
       UNIMCI_check_post__MPI_Type_contiguous(2, MPI_INT, &cont2Int, __FILE__, __LINE__, 8);
       check_unimci_return();

       UNIMCI_check_pre__MPI_Type_contiguous(2, cont2Int, &cont4Int, __FILE__, __LINE__, 9);
       check_unimci_return();
       MPI_Type_contiguous(2, cont2Int, &cont4Int);
       UNIMCI_check_post__MPI_Type_contiguous(2, cont2Int, &cont4Int, __FILE__, __LINE__, 10);
       check_unimci_return();

       if (rank == 0)
       {
            UNIMCI_check_pre__MPI_Send(sendBuf, 1, cont4Int, 1, MSG_TAG_1 ,MPI_COMM_WORLD, __FILE__, __LINE__, 11);
            check_unimci_return();
            MPI_Send(sendBuf, 1, cont4Int, 1, MSG_TAG_1 ,MPI_COMM_WORLD);
            UNIMCI_check_post__MPI_Send(sendBuf, 1, cont4Int, 1, MSG_TAG_1 ,MPI_COMM_WORLD, __FILE__, __LINE__, 12);
            check_unimci_return();

            UNIMCI_check_pre__MPI_Recv(recvBuf, 1, cont4Int, 1, MSG_TAG_2 ,MPI_COMM_WORLD,&status, __FILE__, __LINE__, 13);
            check_unimci_return();
            MPI_Recv(recvBuf, 1, cont4Int, 1, MSG_TAG_2 ,MPI_COMM_WORLD,&status);
            UNIMCI_check_post__MPI_Recv(recvBuf, 1, cont4Int, 1, MSG_TAG_2 ,MPI_COMM_WORLD,&status, __FILE__, __LINE__, 14);
            check_unimci_return();
       }
       if (rank == 1)
       {
            UNIMCI_check_pre__MPI_Recv(recvBuf, 1, cont4Int, 0, MSG_TAG_1 ,MPI_COMM_WORLD,&status, __FILE__, __LINE__, 15);
            check_unimci_return();
            MPI_Recv(recvBuf, 1, cont4Int, 0, MSG_TAG_1 ,MPI_COMM_WORLD,&status);
            UNIMCI_check_post__MPI_Recv(recvBuf, 1, cont4Int, 0, MSG_TAG_1 ,MPI_COMM_WORLD,&status, __FILE__, __LINE__, 16);
            check_unimci_return();

            UNIMCI_check_pre__MPI_Send(sendBuf, 1, cont4Int, 0, MSG_TAG_2 ,MPI_COMM_WORLD, __FILE__, __LINE__, 17);
            check_unimci_return();
            MPI_Send(sendBuf, 1, cont4Int, 0, MSG_TAG_2 ,MPI_COMM_WORLD);
            UNIMCI_check_post__MPI_Send(sendBuf, 1, cont4Int, 0, MSG_TAG_2 ,MPI_COMM_WORLD, __FILE__, __LINE__, 18);
            check_unimci_return();
       }

       UNIMCI_check_pre__MPI_Type_free (&cont4Int, __FILE__, __LINE__, 19);
       check_unimci_return();
       MPI_Type_free (&cont4Int);
       UNIMCI_check_post__MPI_Type_free (&cont4Int, __FILE__, __LINE__, 20);
       check_unimci_return();

       UNIMCI_check_pre__MPI_Type_free (&cont2Int, __FILE__, __LINE__, 21);
       check_unimci_return();
       MPI_Type_free (&cont2Int);
       UNIMCI_check_post__MPI_Type_free (&cont2Int, __FILE__, __LINE__, 22);
       check_unimci_return();
    }

    printf("rank: %d/size: %d buf: %d %d %d %d \n",rank,size,recvBuf[0],recvBuf[1],recvBuf[2],recvBuf[3]);

    UNIMCI_check_pre__MPI_Finalize (__FILE__, __LINE__, 23);
    check_unimci_return();
    MPI_Finalize();
    UNIMCI_check_post__MPI_Finalize (__FILE__, __LINE__, 24);
    check_unimci_return();

    return 0;
}

void check_unimci_return (void)
{
    UNIMCI_MSG *msg;

    while (UNIMCI_has_msg ())
    {
        UNIMCI_pop_msg (&msg);

        print_unimci_message (msg);

        UNIMCI_msg_free (&msg);
    }
}

void print_unimci_message (UNIMCI_MSG *msg)
{
    printf ("%s\n",msg->strText);
}

/*EOF*/
